#include "convolve.h"

/************************ TO BE DONE 1 **************************/
/*
 *	INPUT: 
 *		origImg:		the original image, 
 *		imgWidth:		the width of the image
 *		imgHeight:		the height of the image
 *						the image is arranged such that 
 *						origImg[3*(row*imgWidth+column)+0], 
 *						origImg[3*(row*imgWidth+column)+1], 
 *						origImg[3*(row*imgWidth+column)+2]
 *						are R, G, B values for pixel at (column, row).
 *
 *      kernel:			the 2D filter kernel,
 *		knlWidth:		the width of the kernel
 *		knlHeight:		the height of the kernel
 *      
 *		selection:      a byte array of the same size as the image, 
 *						indicating where in the original image should be filtered, e.g., 
 *						selection[k] == 1 ==> pixel k should be filtered
 *                      selection[k] == 0 ==> pixel k should NOT be filtered
 *                      a special case is selection is a NULL pointer, which means all the pixels should be filtered. 
 *  OUTPUT:
 *		rsltImg:		the filtered image of the same size as original image.
 *						it is a valid pointer ( allocated already ).
 */

void image_filter(double *rsltImg, const double *origImg, const unsigned char *selection, 
			int imgWidth, int imgHeight, 
			const double *kernel, int knlWidth, int knlHeight,
			double scale, double offset)
{
	//copy origImg to rsltImg is NOT the solution, it does nothing!, 
	//you need to figure sth out.
	memcpy(rsltImg,origImg,3*imgWidth*imgHeight*sizeof(double));
	printf("image_filter: to be done in convolve.cpp\n");
}

/************************ END OF TBD 1 **************************/




/************************ TO BE DONE 2 **************************/
/*
 *	INPUT: 
 *      x:				a column index,
 *      y:				a row index,
 *		origImg:		the original image, 
 *		imgWidth:		the width of the image
 *		imgHeight:		the height of the image
 *						the image is arranged such that 
 *						origImg[3*(row*imgWidth+column)+0], 
 *						origImg[3*(row*imgWidth+column)+1], 
 *						origImg[3*(row*imgWidth+column)+2]
 *						are R, G, B values for pixel at (column, row).
 *
 *      kernel:			the 2D filter kernel,
 *		knlWidth:		the width of the kernel
 *		knlHeight:		the height of the kernel
 *      
 *  OUTPUT:
 *		rsltPixel[0], rsltPixel[1], rsltPixel[2]:		
 *						the filtered pixel R, G, B values at row y , column x;
 */

void pixel_filter(double rsltPixel[3], int x, int y, const double *origImg, int imgWidth, int imgHeight, 
			  const double *kernel, int knlWidth, int knlHeight,
			  double scale, double offset)
{
	printf("pixel_filter: to be done in convolve.cpp\n");
}

/************************ END OF TBD 2 **************************/

void image_filter_helper(double *rsltImg, const double *origImg,
				 int imgWidth, int imgHeight, 
				 const double *kernel, int knlWidth, int knlHeight, 
				 double scale, double offset)
{
	int i,j;
	int x,y;

	const int cntKernelX = knlWidth / 2;
	const int cntKernelY = knlHeight / 2;
	
	const int startX = cntKernelX;
	const int startY = cntKernelY;
	const int endX = imgWidth - cntKernelX;
	const int endY = imgHeight - cntKernelY;

	const int dX = 3;
	const int dY = 3*imgWidth;
	const int deltaY = 3*(imgWidth-knlWidth);

	const double *origPixel = origImg -cntKernelX*dX-cntKernelY*dY;			
	double *rsltPixel = rsltImg;

	for (y=0;y<startY;y++)
	{
		for (x=0;x<imgWidth;x++)
		{
			rsltPixel[0] = 0;
			rsltPixel[1] = 0;
			rsltPixel[2] = 0;

			rsltPixel += 3;
			origPixel += 3;
		}
	}

	for (y=startY;y<endY;y++)
	{
		for (x=0;x<startX;x++)
		{
			rsltPixel[0] = 0;
			rsltPixel[1] = 0;
			rsltPixel[2] = 0;

			rsltPixel += 3;
			origPixel += 3;
		}

		for (x=startX;x<endX;x++)
		{
			const double *krnl = kernel;
			const double *data = origPixel;

			rsltPixel[0] = 0;
			rsltPixel[1] = 0;
			rsltPixel[2] = 0;

			for (j=0;j<knlHeight;j++)
			{
				for (i=0;i<knlWidth;i++)
				{
					rsltPixel[0] += (*krnl)*(*data);
					data++;

					rsltPixel[1] += (*krnl)*(*data);
					data++;

					rsltPixel[2] += (*krnl)*(*data);
					data++;

					krnl++;
				}
				data += deltaY;
			}

			rsltPixel[0] *= scale;
			rsltPixel[0] += offset;

			rsltPixel[1] *= scale;
			rsltPixel[1] += offset;

			rsltPixel[2] *= scale;
			rsltPixel[2] += offset;

			rsltPixel += 3;
			origPixel += 3;
		}

		for (x=endX;x<imgWidth;x++)
		{
			rsltPixel[0] = 0;
			rsltPixel[1] = 0;
			rsltPixel[2] = 0;

			rsltPixel += 3;
			origPixel += 3;
		}

	}

	for (y=endY;y<imgHeight;y++)
	{
		for (x=0;x<imgWidth;x++)
		{
			rsltPixel[0] = 0;
			rsltPixel[1] = 0;
			rsltPixel[2] = 0;

			rsltPixel += 3;
			origPixel += 3;
		}
	}
}

void image_filter_helper(double *rsltImg, const double *origImg, const unsigned char *selection, 
			  int imgWidth, int imgHeight, 
			  const double *kernel, int knlWidth, int knlHeight,
			  double scale, double offset)
{
	int i,j;
	int x,y;

	const int cntKernelX = knlWidth / 2;
	const int cntKernelY = knlHeight / 2;
	
	const int startX = cntKernelX;
	const int startY = cntKernelY;
	const int endX = imgWidth - cntKernelX;
	const int endY = imgHeight - cntKernelY;

	const int dX = 3;
	const int dY = 3*imgWidth;
	const int deltaY = 3*(imgWidth-knlWidth);
	const int kernelOffset = cntKernelX*dX+cntKernelY*dY;

	const unsigned char *sel = selection;
	const double *origPixel = origImg;
	double *rsltPixel = rsltImg;

	for (y=0;y<startY;y++)
	{
		for (x=0;x<imgWidth;x++)
		{
			rsltPixel[0] = 0;
			rsltPixel[1] = 0;
			rsltPixel[2] = 0;

			rsltPixel += 3;
			origPixel += 3;
			sel++;
		}
	}

	for (y=startY;y<endY;y++)
	{
		for (x=0;x<startX;x++)
		{
			rsltPixel[0] = 0;
			rsltPixel[1] = 0;
			rsltPixel[2] = 0;

			rsltPixel += 3;
			origPixel += 3;
			sel++;
		}

		for (x=startX;x<endX;x++)
		{
			if (*sel)
			{
				const double *krnl = kernel;
				const double *data = origPixel-kernelOffset;

				rsltPixel[0] = 0;
				rsltPixel[1] = 0;
				rsltPixel[2] = 0;

				for (j=0;j<knlHeight;j++)
				{
					for (i=0;i<knlWidth;i++)
					{
						rsltPixel[0] += (*krnl)*(*data);
						data++;

						rsltPixel[1] += (*krnl)*(*data);
						data++;

						rsltPixel[2] += (*krnl)*(*data);
						data++;

						krnl++;
					}
					data += deltaY;
				}

				rsltPixel[0] *= scale;
				rsltPixel[0] += offset;

				rsltPixel[1] *= scale;
				rsltPixel[1] += offset;

				rsltPixel[2] *= scale;
				rsltPixel[2] += offset;
			}
			else 
			{
				rsltPixel[0] = origPixel[0];
				rsltPixel[1] = origPixel[1];
				rsltPixel[2] = origPixel[2];
			}

			rsltPixel += 3;
			origPixel += 3;
			sel++;
		}

		for (x=endX;x<imgWidth;x++)
		{
			rsltPixel[0] = 0;
			rsltPixel[1] = 0;
			rsltPixel[2] = 0;

			rsltPixel += 3;
			origPixel += 3;
			sel++;
		}

	}

	for (y=endY;y<imgHeight;y++)
	{
		for (x=0;x<imgWidth;x++)
		{
			rsltPixel[0] = 0;
			rsltPixel[1] = 0;
			rsltPixel[2] = 0;

			rsltPixel += 3;
			origPixel += 3;
			sel++;
		}
	}
}

